import isEqual from 'fast-deep-equal';
import { useCallback, useEffect, useImperativeHandle, useRef, useState, } from 'react';
import usePrevious from '@restart/hooks/usePrevious';
import { clearTypeahead, clickOrFocusInput, getInitialState, hideMenu as _hideMenu, toggleMenu as _toggleMenu, useTypeaheadState, } from './TypeaheadState';
import useValidateProps from './useValidateProps';
import { addCustomOption, defaultFilterBy, defaultSelectHint, getHintText, getInputProps, getInputText, getIsOnlyResult, getMenuProps, getOptionLabel, getOptionProperty, getStringLabelKey, getTruncatedOptions, getUpdatedActiveIndex, isFunction, isShown, isString, noop, uniqueId, validateSelectedPropChange, } from '../utils';
import { DEFAULT_LABELKEY } from '../constants';
const defaultProps = {
    allowNew: false,
    autoFocus: false,
    caseSensitive: false,
    defaultInputValue: '',
    defaultOpen: false,
    defaultSelected: [],
    filterBy: [],
    highlightOnlyResult: false,
    ignoreDiacritics: true,
    inputProps: {},
    labelKey: DEFAULT_LABELKEY,
    maxResults: 100,
    minLength: 0,
    multiple: false,
    onBlur: noop,
    onFocus: noop,
    onInputChange: noop,
    onKeyDown: noop,
    onMenuToggle: noop,
    onPaginate: noop,
    paginate: true,
};
function triggerInputChange(input, value) {
    const inputValue = Object.getOwnPropertyDescriptor(window.HTMLInputElement.prototype, 'value');
    inputValue && inputValue.set && inputValue.set.call(input, value);
    const e = new Event('input', { bubbles: true });
    input.dispatchEvent(e);
}
function useDidUpdate(props, state, setState) {
    const prevProps = usePrevious(props);
    const isInitialRender = useRef(true);
    useEffect(() => {
        if (isInitialRender.current) {
            isInitialRender.current = false;
            return;
        }
        validateSelectedPropChange(props.selected, prevProps?.selected);
        if (props.selected && !isEqual(props.selected, state.selected)) {
            setState({ selected: props.selected });
            if (!props.multiple) {
                setState({
                    text: props.selected.length
                        ? getOptionLabel(props.selected[0], props.labelKey)
                        : '',
                });
            }
        }
    }, [
        props.labelKey,
        props.multiple,
        props.selected,
        prevProps?.selected,
        setState,
        state.selected,
    ]);
}
function useOnMenuToggle(isMenuShown, onMenuToggle) {
    const isInitialRender = useRef(true);
    useEffect(() => {
        if (isInitialRender.current) {
            isInitialRender.current = false;
            return;
        }
        onMenuToggle?.(isMenuShown);
    }, [isMenuShown, onMenuToggle]);
}
function useTypeahead({ onChange, ...partialProps }, ref) {
    const props = {
        ...defaultProps,
        ...partialProps,
    };
    const [state, setState] = useTypeaheadState(props);
    const [inputNode, setInputNode] = useState(null);
    const mergedPropsAndState = { ...props, ...state };
    const { filterBy, labelKey, options, paginate } = mergedPropsAndState;
    const isMenuShown = isShown(mergedPropsAndState);
    const items = [];
    const hintText = getHintText({ ...mergedPropsAndState, isMenuShown });
    useValidateProps(props);
    useDidUpdate(props, state, setState);
    useOnMenuToggle(isMenuShown, props.onMenuToggle);
    function hideMenu() {
        setState(_hideMenu);
    }
    function blur() {
        inputNode?.blur();
        hideMenu();
    }
    function clear() {
        setState(clearTypeahead);
    }
    const focus = useCallback(() => {
        inputNode?.focus();
    }, [inputNode]);
    function getInput() {
        return inputNode;
    }
    function toggleMenu() {
        setState(_toggleMenu);
    }
    useImperativeHandle(ref, () => ({
        blur,
        clear,
        focus,
        getInput,
        hideMenu,
        toggleMenu,
    }));
    useEffect(() => {
        props.autoFocus && focus();
    }, [props.autoFocus, focus]);
    let results = [];
    if (isMenuShown) {
        const cb = (isFunction(filterBy) ? filterBy : defaultFilterBy);
        results = options.filter((option) => cb(option, mergedPropsAndState));
        const shouldPaginate = paginate && results.length > state.shownResults;
        results = getTruncatedOptions(results, state.shownResults);
        if (addCustomOption(results, mergedPropsAndState)) {
            results.push({
                customOption: true,
                [getStringLabelKey(labelKey)]: state.text,
            });
        }
        if (shouldPaginate) {
            results.push({
                [getStringLabelKey(labelKey)]: '',
                paginationOption: true,
            });
        }
    }
    const isOnlyResult = getIsOnlyResult({ ...props, results });
    function setItem(item, position) {
        items[position] = item;
    }
    function onActiveIndexChange(index) {
        setState({
            activeIndex: index,
            activeItem: index >= 0 ? items[index] : undefined,
        });
    }
    function onBlur(e) {
        setState({
            ..._hideMenu(state, props),
            isFocused: false,
        }, () => props.onBlur(e));
    }
    function onSelect(newSelected) {
        onChange && onChange(newSelected);
    }
    function onClear() {
        inputNode && triggerInputChange(inputNode, '');
        setState(clearTypeahead, () => {
            if (props.multiple) {
                onSelect([]);
            }
        });
    }
    function onClick(e) {
        const handleClick = props.inputProps?.onClick;
        setState(clickOrFocusInput, () => handleClick && handleClick(e));
    }
    function onFocus(e) {
        setState(clickOrFocusInput, () => props.onFocus(e));
    }
    function onInitialItemChange(initialItem) {
        if (!isEqual(initialItem, state.initialItem)) {
            setState({ initialItem });
        }
    }
    function onInputChange(e) {
        const { value } = e.currentTarget;
        const shouldClearSelections = state.selected.length && !props.multiple;
        const { activeIndex, activeItem, shownResults } = getInitialState(props);
        setState({
            activeIndex,
            activeItem,
            selected: shouldClearSelections ? [] : state.selected,
            showMenu: true,
            shownResults,
            text: value,
        });
        props.onInputChange(value, e);
        shouldClearSelections && onSelect([]);
    }
    function onAdd(option) {
        let selected;
        let selection = option;
        let text = '';
        if (!isString(selection) && selection.customOption) {
            selection = { ...selection, id: uniqueId('new-id-') };
        }
        if (props.multiple) {
            selected = [...state.selected, selection];
        }
        else {
            selected = [selection];
            text = getOptionLabel(selection, props.labelKey);
        }
        setState({
            ..._hideMenu(state, props),
            initialItem: selection,
            selected,
            text,
        }, (newState) => onSelect(newState.selected));
    }
    function onPaginate(e) {
        setState((currentState) => ({
            shownResults: currentState.shownResults + props.maxResults,
        }), (newState) => props.onPaginate(e, newState.shownResults));
    }
    function onMenuItemSelect(option, e) {
        if (getOptionProperty(option, 'paginationOption')) {
            onPaginate(e);
        }
        else {
            onAdd(option);
        }
    }
    function onKeyDown(e) {
        props.onKeyDown(e);
        if (!isMenuShown) {
            if (e.key === 'ArrowUp' || e.key === 'ArrowDown') {
                setState({ showMenu: true });
            }
            return;
        }
        switch (e.key) {
            case 'ArrowUp':
            case 'ArrowDown':
                e.preventDefault();
                onActiveIndexChange(getUpdatedActiveIndex(state.activeIndex, e.key, items));
                break;
            case 'Enter':
                e.preventDefault();
                state.activeItem && onMenuItemSelect(state.activeItem, e);
                break;
            case 'Escape':
            case 'Tab':
                hideMenu();
                break;
            default:
                break;
        }
        if (!state.initialItem) {
            return;
        }
        const addOnlyResult = e.key === 'Enter' && isOnlyResult;
        const shouldSelectHint = hintText && defaultSelectHint(e, props.selectHint);
        if (addOnlyResult || shouldSelectHint) {
            onAdd(state.initialItem);
        }
    }
    function onRemove(selection) {
        const selected = state.selected.filter((option) => !isEqual(option, selection));
        focus();
        setState({
            ..._hideMenu(state, props),
            selected,
        }, () => onSelect(selected));
    }
    useEffect(() => {
        if (!(props.allowNew || results.length)) {
            onInitialItemChange();
        }
    });
    const context = {
        activeIndex: state.activeIndex,
        hintText,
        id: props.id,
        initialItem: state.initialItem,
        inputNode,
        isOnlyResult,
        onAdd,
        onInitialItemChange,
        onMenuItemClick: onMenuItemSelect,
        setItem,
    };
    return {
        ...mergedPropsAndState,
        context,
        getInputProps: getInputProps({
            activeIndex: state.activeIndex,
            id: props.id,
            inputRef: setInputNode,
            isFocused: state.isFocused,
            isMenuShown,
            multiple: props.multiple,
            onBlur,
            onChange: onInputChange,
            onClick,
            onFocus,
            onKeyDown,
            value: getInputText(mergedPropsAndState),
        }),
        getMenuProps: getMenuProps({
            id: props.id,
        }),
        hideMenu,
        inputNode,
        isMenuShown,
        onClear,
        onHide: hideMenu,
        onRemove,
        results,
        toggleMenu,
    };
}
export default useTypeahead;
